[![Scand Storm Petrel Generator](https://raw.githubusercontent.com/Scandltd/storm-petrel/main/generator/assets/logo-128x128-transparent.png)](https://scand.com/products/storm-petrel-expected-baselines-rewriter)
# Scand Storm Petrel Generator
* [Overview](#overview)
* [Primary Use Cases](#primary-use-cases)
    * [Enabler: Auto-Generate `CalculatorTestStormPetrel.AddTestStormPetrel`](#enabler-auto-generate-calculatorteststormpetreladdteststormpetrel)
    * [Overwrite `CalculatorTest.AddTest` expected baseline](#overwrite-calculatortestaddtest-expected-baseline)
    * [Overwrite `CalculatorTestTheory.AddTestGetExpected` expected baseline method](#overwrite-calculatortesttheoryaddtestgetexpected-expected-baseline-method)
    * [Use Case Variations](#use-case-variations)
        * [Expected variable expression is static property assignment](#expected-variable-expression-is-static-property-assignment)
        * [Expected expression is test method argument what comes from](#expected-expression-is-test-method-argument-what-comes-from)
            * [Test case attribute](#test-case-attribute)
            * [Test data source attribute](#test-data-source-attribute)
        * [Snapshots of Snapshot Testing](#snapshots-of-snapshot-testing)
        * [Expected expression decorators](#expected-expression-decorators)
            * [Collection initializer](#collection-initializer)
            * [Implicit object creation](#implicit-object-creation)
            * [Removal of property or field assignments](#removal-of-property-or-field-assignments)
            * [Verbatim strings or custom literal values](#verbatim-strings-or-custom-literal-values)
        * [Expected expression is inlined within an assertion](#expected-expression-is-inlined-within-an-assertion)
            * [FluentAssertions expressions](#fluentassertions-expressions)
            * [xUnit assertions](#xunit-assertions)
            * [NUnit assertions](#nunit-assertions)
            * [MSTest assertions](#mstest-assertions)
            * [Shouldly expressions](#shouldly-expressions)
            * [Custom assertions that replicate the signatures above](#custom-assertions-that-replicate-the-signatures-above)
        * [Framework and library specific examples](#framework-and-library-specific-examples)
            * [ASP.NET Core](#aspnet-core)
            * [Exceptions](#exceptions)
            * [Windows Forms .NET](#windows-forms-net)
            * [WPF .NET](#wpf-net)
            * [AutoFixture](#autofixture)
            * [Playwright](#playwright)
            * [Other examples](#other-examples)
        * [Actual value is built via mocks or callbacks or other algorithms](#actual-value-is-built-via-mocks-or-callbacks-or-other-algorithms)
* [Getting Started](#getting-started)
* [Configuration](#configuration)
    * [Optional Visual Studio Extension](#optional-visual-studio-extension)
    * [Optional JSON File](#optional-json-file)
    * [Optional Environment Variable](#optional-environment-variable)
* [Supported Software](#supported-software)
    * [Test Frameworks](#test-frameworks)
    * [.NET Versions](#net-versions)
    * [Operating Systems and IDEs](#operating-systems-and-ides)
* [CHANGELOG](#changelog)
* [FAQs](#faqs)
* [References](#references)

## Overview
[![NuGet Version](http://img.shields.io/nuget/v/Scand.StormPetrel.Generator.svg?style=flat)](https://www.nuget.org/packages/Scand.StormPetrel.Generator)

.NET Incremental Generator that creates modified copies of unit and/or integration tests to update expected baselines in original tests, automating baseline creation and accelerating test development.

[![Primary Use Case](https://raw.githubusercontent.com/Scandltd/storm-petrel/main/generator/assets/primary-use-case.gif)](https://raw.githubusercontent.com/Scandltd/storm-petrel/main/generator/assets/primary-use-case.gif)

## Primary Use Cases

### Enabler: Auto-Generate `CalculatorTestStormPetrel.AddTestStormPetrel`

###### Given
`Calculator` class with a bug introduced by the `buggyDelta` variable:
```csharp
public class Calculator
{
    public static AddResult Add(int a, int b)
    {
        var buggyDelta = 1;
        var result = a + b + buggyDelta;
        return new AddResult
        {
            Value = result,
            ValueAsHexString = "0x" + result.ToString("x"),
        };
    }
}

public class AddResult
{
    public int Value { get; set; }
    public string ValueAsHexString { get; set; } = string.Empty;
}
```

And its corresponding test with an expected baseline that matches the `Calculator.Add` buggy behavior:

```csharp
public class CalculatorTest
{
    [Xunit.Fact]
    public void AddTest()
    {
        //Arrange
        // incorrect `expected` baseline value will be overwritten with correct `actual` value
        // after manual execution of auto-generated AddTestStormPetrel test.
        var expected = new AddResult
        {
            Value = 5, //incorrect value example
            ValueAsHexString = "0x5"
        };

        //Act
        var actual = Calculator.Add(2, 2);

        //Assert
        actual.Should().BeEquivalentTo(expected);
    }
}
```

##### When
The developer configures the test project with StormPetrel.Generator as per the [Getting Started](#getting-started).

##### Then
A new test method, `CalculatorTestStormPetrel.AddTestStormPetrel`, is generated. This method is a specially modified copy of the original CalculatorTest.AddTest to overwrite its expected baseline.

### Overwrite `CalculatorTest.AddTest` expected baseline
##### Given
`CalculatorTestStormPetrel.AddTestStormPetrel` is auto-generated after enabling Storm Petrel functionality.
##### When
The developer fixes `buggyDelta` to `0`
and executes `CalculatorTestStormPetrel.AddTestStormPetrel` test.
##### Then
`CalculatorTest.AddTest` code is populated with correct expected baseline value, i.e. its code becomes
```csharp
public class CalculatorTest
{
    [Xunit.Fact]
    public void AddTest()
    {
        //Arrange
        // incorrect `expected` baseline value will be overwritten with correct `actual` value
        // after manual execution of auto-generated AddTestStormPetrel test.
        var expected = new AddResult
        {
            Value = 4,
            ValueAsHexString = "0x4"
        };

        //Act
        var actual = Calculator.Add(2, 2);

        //Assert
        actual.Should().BeEquivalentTo(expected);
    }
}
```
##### So that
The developer should only review expected baseline changes (no manual modification) what typically saves development time.

### Overwrite `CalculatorTestTheory.AddTestGetExpected` expected baseline method
##### Given
`CalculatorTestTheory.AddTest` test below with

* multiple use cases;
* AddTestGetExpected static method call returning incorrect expected baselines based on arguments;
* AddTestGetExpected method possible variations per its comments.
```csharp
public class CalculatorTestTheory
{
    [Xunit.Theory]
    [Xunit.InlineData(1, 5)]
    [Xunit.InlineData(2, 2)]
    [Xunit.InlineData(2, 3)]
    public void AddTest(int a, int b)
    {
        //Arrange
        var expected = AddTestGetExpected(a, b);

        //Act
        var actual = Calculator.Add(a, b);

        //Assert
        actual.Should().BeEquivalentTo(expected);
    }

    /// <summary>
    /// Possible variations of AddTestGetExpected static method are:
    /// - Method body may have pattern matches within pattern matches.
    /// - Method body may have `switch` and/or `if` expressions with return statements returning expected baselines.
    /// - The method may be placed in another class and/or file.
    /// </summary>
    private static AddResult AddTestGetExpected(int a, int b) => (a, b) switch
    {
        (1, 5) => new AddResult(), // should be overwritten with correct expected baseline after
                                   // CalculatorTestTheoryStormPetrel.AddTestStormPetrel execution
        (2, 2) => new AddResult(),
        (2, 3) => new AddResult(),
        _ => throw new InvalidOperationException(),
    };
}
```
##### When
The developer executes `CalculatorTestTheoryStormPetrel.AddTestStormPetrel` test.
##### Then
`CalculatorTestTheory.AddTestGetExpected` code is populated with correct expected baseline values, i.e. its code becomes
```csharp
private static AddResult AddTestGetExpected(int a, int b) => (a, b) switch
{
    (1, 5) => new AddResult
    {
        Value = 6,
        ValueAsHexString = "0x6"
    }, //should be overwritten with correct expected baseline
    (2, 2) => new AddResult
    {
        Value = 4,
        ValueAsHexString = "0x4"
    },
    (2, 3) => new AddResult
    {
        Value = 5,
        ValueAsHexString = "0x5"
    },
    _ => throw new InvalidOperationException(),
};
```
##### So that
The developer should only review expected baseline changes (no manual modification) what typically saves development time.

### Use Case Variations

#### Expected variable expression is static property assignment
See [PropertyTest](https://github.com/Scandltd/storm-petrel/blob/main/generator/Test.Integration.XUnit/PropertyTest.cs) for more details.

#### Expected expression is test method argument what comes from

##### Test case attribute
Supported attributes are xUnit InlineData, NUnit TestCase, MSTest DataRow. See [AttributesTest](https://github.com/Scandltd/storm-petrel/blob/main/generator/Test.Integration.XUnit/AttributesTest.cs) for more details.

##### Test data source attribute
Supported attributes are xUnit MemberData or ClassData, NUnit TestCaseSource, MSTest DynamicData. See [TestCaseSourceMemberDataTest](https://github.com/Scandltd/storm-petrel/blob/main/generator/Test.Integration.XUnit/TestCaseSourceMemberDataTest.cs), [TestCaseSourceClassDataTest](https://github.com/Scandltd/storm-petrel/blob/main/generator/Test.Integration.XUnit/TestCaseSourceClass/TestCaseSourceClassDataTest.cs), [NUnit TestCaseSourceTest](https://github.com/Scandltd/storm-petrel/blob/main/generator/Test.Integration.NUnit/TestCaseSourceTest.cs), [MSTest TestCaseSourceTest](https://github.com/Scandltd/storm-petrel/blob/main/generator/Test.Integration.MSTest/TestCaseSourceTest.cs) for more details.
Known limitations:
* The attributes should be configured against data source methods or properties, not fields.

#### Snapshots of Snapshot Testing
HTML, JSON, binary or whatever expected snapshots can be hardcoded in tests code. See couple examples in [SnapshotTest](https://github.com/Scandltd/storm-petrel/blob/main/generator/Test.Integration.XUnit/SnapshotTest.cs).

#### Expected expression decorators
Refer to the classes in the `Scand.StormPetrel.Generator.Utils` namespace and their usage in the files of the [Test.Integration.Generator.Utils.XUnit](https://github.com/Scandltd/storm-petrel/blob/main/generator/Test.Integration.Generator.Utils.XUnit/Test.Integration.Generator.Utils.XUnit.csproj) project to decorate the expressions. The example includes:
* [CustomDumperDecorator](https://github.com/Scandltd/storm-petrel/blob/main/generator/Test.Integration.Generator.Utils.XUnit/CustomDumperDecorator.cs). Replaces `DateTime.ParseExact(...)` with `new DateTime(...)` or a constant value.
* [DumperFactory](https://github.com/Scandltd/storm-petrel/blob/main/generator/Test.Integration.Generator.Utils.XUnit/DumperFactory.cs). Combines implementations of `IGeneratorDumper` to have custom C# representations of expected baselines. It demonstrates how to remove redundant assignments and apply constant expressions, verbatim strings, implicit object creation, collection expressions and other C# syntax decorations to the baselines.

Built-in decorators include

##### Collection initializer
Replaces C# syntax nodes (e.g., array creation) with collection expressions.

##### Implicit object creation
Replaces C# constructor calls with implicit object creation syntax.

##### Removal of property or field assignments
Removes property of field assignments according to the configuration.

##### Verbatim strings or custom literal values
Replaces regular strings with verbatim strings, regular literal values with custom literal values (e.g., regular numbers with hex numbers) according to the configuration.

#### Expected expression is inlined within an assertion

With StormPetrel, the expected variable can be entirely omitted. Instead, you can inline expected expressions of the following kinds:
- Anonymous object creation expressions;
- Array creation expressions;
- Collection expressions;
- Implicit array creation expressions;
- Implicit object creation expressions;
- Literal expressions (e.g., numeric or string literals);
- Object creation expressions;
- Tuple expressions.

Refer to the sections below for more details about supported assertions and examples of expected expressions.

##### FluentAssertions expressions
Supported expressions include `actual.Should().Be(123);`, `actual.SomePropertyOrMethodCall.Should().Be("string value");`, and `actual.Should().BeEquivalentTo(new MyClass{...});` according to [FluentAssertions](https://fluentassertions.com/) documentation. See more examples in [NoExpectedVarTest](https://github.com/Scandltd/storm-petrel/blob/main/generator/Test.Integration.XUnit/NoExpectedVarTest.cs) and [NoExpectedVarExpressionKindsTest](https://github.com/Scandltd/storm-petrel/blob/main/generator/Test.Integration.XUnit/NoExpectedVarExpressionKindsTest.cs).

##### xUnit assertions
Supported expressions include `Assert.Equal(123, actual);`, `Assert.StrictEqual("string value", actual);`, and `Assert.Equivalent("string value", actual);`. See more examples in [NoExpectedVarAssertTest](https://github.com/Scandltd/storm-petrel/blob/main/generator/Test.Integration.XUnit/NoExpectedVarAssertTest.cs).

##### NUnit assertions
Supported expressions include `Assert.That(actual, Is.EqualTo(123));` and `Assert.That(actual, Is.EquivalentTo(new MyClass{...}));`. See more examples in [NoExpectedVarAssertThatTest](https://github.com/Scandltd/storm-petrel/blob/main/generator/Test.Integration.NUnit/NoExpectedVarAssertThatTest.cs).

##### MSTest assertions
Supported expressions include `Assert.AreEqual(123, actual);`. See more examples in [NoExpectedVarAssertTest](https://github.com/Scandltd/storm-petrel/blob/main/generator/Test.Integration.MSTest/NoExpectedVarAssertTest.cs).

##### Shouldly expressions
Supported expressions include `actual.ShouldBe(123);` according to [Shouldly](https://docs.shouldly.org/) documentation. See more examples in [NoExpectedVarShouldlyTest](https://github.com/Scandltd/storm-petrel/blob/main/generator/Test.Integration.XUnit/NoExpectedVarShouldlyTest.cs).

##### Custom assertions that replicate the signatures above
You can also implement and use custom assertions. Ensure they replicate the signatures of the assertions mentioned above so that StormPetrel can detect them effectively.

#### Framework and library specific examples

##### ASP.NET Core
See test examples in [AspNetTest](https://github.com/Scandltd/storm-petrel/blob/main/generator/Test.Integration.WebApi.XUnit/AspNetTest.cs).

##### Exceptions
See test examples in [ExceptionTest](https://github.com/Scandltd/storm-petrel/blob/main/generator/Test.Integration.XUnit/ExceptionTest.cs).

##### Windows Forms .NET
See test examples in [WinFormsTest](https://github.com/Scandltd/storm-petrel/blob/main/generator/../file-snapshot-infrastructure/Test.Integration.WinFormsAppTest/WinFormsTest.cs) of [Scand.StormPetrel.FileSnapshotInfrastructure](https://github.com/Scandltd/storm-petrel/blob/main/generator/../file-snapshot-infrastructure/README.md).

##### WPF .NET
See examples in [Scand.StormPetrel.FileSnapshotInfrastructure](https://github.com/Scandltd/storm-petrel/blob/main/generator/../file-snapshot-infrastructure/README.md) tests:
* [WpfTest](https://github.com/Scandltd/storm-petrel/blob/main/generator/../file-snapshot-infrastructure/Test.Integration.WpfAppTest/WpfTest.cs) for regular xUnit attributes, but more complex test implementation.
* [WpfCustomAttributeTest](https://github.com/Scandltd/storm-petrel/blob/main/generator/../file-snapshot-infrastructure/Test.Integration.WpfAppTest/WpfCustomAttributeTest.cs) for additional configuration of [Xunit.StaFact](https://www.nuget.org/packages/Xunit.StaFact) NuGet package attributes, but simpler test implementation.

##### AutoFixture
See [AutoFixtureTest](https://github.com/Scandltd/storm-petrel/blob/main/generator/Test.Integration.XUnit/AutoFixtureTest.cs) for [AutoFixture](https://github.com/AutoFixture/AutoFixture) examples. Find `AutoFixture.Xunit2.InlineAutoDataAttribute` in the [build](https://github.com/Scandltd/storm-petrel/blob/main/generator/../build/build.ps1) script to see AutoFixture attributes configuration example for the environment variable what can be configured at process, user or machine level.

##### Playwright
See [PlaywrightTest](https://github.com/Scandltd/storm-petrel/blob/main/generator/../file-snapshot-infrastructure/Test.Integration.PlaywrightTest) for [Playwright](https://playwright.dev/) test project configuration and examples. This project also showcases the Storm Petrel setup, enabling both traditional tests and file snapshot tests within the same project and even the same test class.

##### Other examples
The preceding sections highlight key test examples. While not exhaustive, Storm Petrel can also provide compatibility for the following frameworks/libraries in a similar manner:
* [Aspose](https://www.aspose.com/).
* [Avalonia UIs](https://avaloniaui.net/).
* [Blazor Components](https://docs.microsoft.com/en-us/aspnet/core/blazor/#components).
* [FreeSpire.PDF](https://www.e-iceblue.com/Introduce/free-pdf-component.html).
* [iText](https://itextpdf.com/) (aka iTextSharp).
* [Newtonsoft.Json](https://github.com/JamesNK/Newtonsoft.Json).
* [Puppeteer Sharp](https://github.com/hardkoded/puppeteer-sharp).
* [Selenium](https://www.selenium.dev/).
* [Syncfusion Essential Document Processing](https://help.syncfusion.com/document-processing/introduction).

#### Actual value is built via mocks or callbacks or other algorithms
Storm Petrel injects baseline update logic **immediately after** the last `actual` or `expected` value assignment.

When actual values are generated via:
* Mocks (e.g., Moq/NSubstitute)
* Callbacks
* Complex algorithms

The baseline rewriting behavior may require test adjustments for reliable updates.
See [BuildActualTest](https://github.com/Scandltd/storm-petrel/blob/main/generator/Test.Integration.XUnit/BuildActualTest.cs) for:
* ✅ Working patterns for baseline updates
* ⚠️ Edge cases requiring modifications

## Getting Started
To utilize the StormPetrel tests, add the following NuGet Package references to your test project:
* Scand.StormPetrel.Generator.
* Object to C# code dumper. Typically, it represents `actual` test instance as C# code. See `DumperExpression` in [Configuration](#configuration) for more details.
    * **Option A**. [VarDump](https://www.nuget.org/packages/VarDump). Must be referenced in the case of StormPetrel.Generator default configuration (no `appsettings.StormPetrel.json` file in the test project). May be additionally configured.
    * **Option B**. [ObjectDumper.NET](https://github.com/thomasgalliker/ObjectDumper). May be referenced and configured.
    * **Option C**. Custom implementation of `IGeneratorDumper` interface. May be developed and configured.
    * **Option D**. [Scand.StormPetrel.FileSnapshotInfrastructure](https://www.nuget.org/packages/Scand.StormPetrel.FileSnapshotInfrastructure). Typically, it represents the `actual` test instance as a checksum and writes the instance bytes to a snapshot file in the `IGeneratorRewriter` implementation. It may be referenced and configured according to its [settings](https://github.com/Scandltd/storm-petrel/blob/main/generator/../file-snapshot-infrastructure/README.md#getting-started).

## Configuration

### Optional Visual Studio Extension

Use [Visual Studio extension](https://marketplace.visualstudio.com/items?itemName=ScandDevteam.Scand-StormPetrel-Extension) for configuring Storm Petrel. It provides a UI to set up test project settings (adds necessary NuGet packages and `appsettings.StormPetrel.json` file to the project), enabling automatic rewriting of expected baselines.

**Alternative:** Manually install packages and configure settings.

### Optional JSON File

The StormPetrel Generator introduces several interfaces and classes to the Scand.StormPetrel.Generator.TargetProject namespace of the test project. These can be utilized alongside an optional JSON file to customize the rewriting of expected baselines. Key interfaces and classes include:
* `IGenerator`, `Generator`;
* `IGeneratorBackuper`, `GeneratorBackuper`;
* `IGeneratorDumper`, `GeneratorDumper`;
* `IGeneratorRewriter`, `GeneratorRewriter`.

Optionally `appsettings.StormPetrel.json` file (its `Build Action` should be `C# analyzer additional file`) can be added to a test project to configure Storm Petrel functionality.
The file changes are applied `on the fly` and can have the following settings:
```jsonc
{
  "$schema": "https://raw.githubusercontent.com/Scandltd/storm-petrel/main/generator/assets/appsettings.StormPetrel.Schema.json", // [optional] string, path to json schema.
  "TargetProjectGeneratorExpression": "...", // [optional] string, configures the default `Generator`. An expression for the `IGenerator` instance.
  "GeneratorConfig":             // [optional] object to configure `Generator` behavior.
  {
    "BackuperExpression": "...", // [optional] string, instantiates `GeneratorBackuper` by default. An expression for the `IGeneratorBackuper` instance. Set to 'null' to skip creating backup files.
    "DumperExpression": "...",   // [optional] string, instantiates `GeneratorDumper` by default. An expression for the `IGeneratorDumper` instance. `GeneratorDumper` references [VarDump](https://www.nuget.org/packages/VarDump) stuff. Use
                                 // - "new Scand.StormPetrel.Generator.TargetProject.GeneratorDumper(CustomCSharpDumperProvider.GetCSharpDumper())" to have `VarDump` with custom options. Need to implement `CustomCSharpDumperProvider.GetCSharpDumper()` method in this case.
                                 // - "new Scand.StormPetrel.Generator.TargetProject.GeneratorObjectDumper()" expression for `GeneratorObjectDumper` instance which references [ObjectDumper.NET](https://github.com/thomasgalliker/ObjectDumper) stuff.
                                 // - "new Scand.StormPetrel.Generator.TargetProject.GeneratorObjectDumper(CustomOptionsProvider.GetDumpOptions())" to have `ObjectDumper.NET` with custom options. Need to implement `CustomOptionsProvider.GetDumpOptions()` method in this case.
                                 // - "new CustomClassImplementingIGeneratorDumper()" or similar expression to have totally custom implementation of dumping of an instance to C# code.
    "RewriterExpression": "..."  // [optional] string, instantiates `GeneratorRewriter` by default. An expression for the `IGeneratorRewriter` instance.
  },
  "IsDisabled": false,           // [optional] boolean, false is by default. Indicates whether the generator should create 'StormPetrel' classes.
                                 // Even if set to 'false', the generator still adds classes like 'IGeneratorDumper', 'GeneratorDumper' to avoid test project compilation failures
                                 // in the case when custom classes uses them.
  "IgnoreFilePathRegex": "...",  // [optional] string, empty by default. Regular Expression to exclude certain paths from 'StormPetrel' class generation.
  "IgnoreInvocationExpressionRegex": "...",  // [optional] string, empty by default. Regular Expression to detect invocation expressions to not execute StormPetrel rewriting for.
                                 // The property can be utilized in the case of custom IGeneratorRewriter implementations (e.g. when expected baseline is not stored in C# code but binary file as in File Snapshot Testing approach).
  "Serilog": "...",              // [optional] Logging configuration using Serilog (https://github.com/serilog/serilog-settings-configuration?tab=readme-ov-file#serilogsettingsconfiguration--).
                                 // Defaults to logging warnings to the test project's Logs folder. Set to 'null' to disable logging.
                                 // Use the '{StormPetrelRootPath}' token to indicate the target test project root path.
  "TestVariablePairConfigs": [   // [optional] array of objects. Configures naming pairs for actual/expected variables to generate correct expected baselines.
    {
      "ActualVarNameTokenRegex": "[Aa]{1}ctual",     // Default configuration object. Assumes variable pair names like (expected, actual), (myExpected, myActual), (expectedOne, actualOne), (ExpectedTwo, ActualTwo), etc.
      "ExpectedVarNameTokenRegex": "[Ee]{1}xpected", // Corresponds to the `ActualVarNameTokenRegex` for pairing.
    },
    {
      "ActualVarNameTokenRegex": "[Aa]{1}ctual",     // One more default configuration object.
      "ExpectedVarNameTokenRegex": null,             // A `null` value indicates that the expected variable may be missing.
                                                     // In such cases, StormPetrel should analyze the assertion expressions to rewrite their `expected` segments with actual values.
    }
  ]
}
```

### Optional Environment Variable

To enable custom test attribute support in Scand.StormPetrel.Generator, you may set the `SCAND_STORM_PETREL_GENERATOR_CONFIG` environment variable. This JSON configuration will be read via `Environment.GetEnvironmentVariable()`  during incremental generation. Configuration format:
```jsonc
{
    "CustomTestAttributes": [                       // Optional array of custom test attribute configurations
        {
            "TestFrameworkKindName": "XUnit",       // Required: "XUnit", "NUnit", or "MSTest" (case-insensitive)
            "FullName": "Xunit.UIFactAttribute",    // Required: Full namespace-qualified attribute name (case-sensitive)
            "KindName":"Test",                      // Required: "Test", "TestCase", or "TestCaseSource" (case-insensitive)
            "XUnitTestCaseSourceKindName": "..."    // Required for xUnit TestCaseSource: "MemberData" or "ClassData"
        }
    ]
}
```
Configure the variable at process, user or machine level. IDE and/or incremental generator dotnet processes restart may be required to apply the configuration change.
See:
* Process-level configuration example in [build.ps1](https://github.com/Scandltd/storm-petrel/blob/main/generator/../build/build.ps1).
* Custom attributes examples in [MSTest CustomAttributeTest](https://github.com/Scandltd/storm-petrel/blob/main/generator/Test.Integration.MSTest/CustomAttributes/CustomAttributeTest.cs), [NUnit CustomAttributeTest](https://github.com/Scandltd/storm-petrel/blob/main/generator/Test.Integration.NUnit/CustomAttributes/CustomAttributeTest.cs), [XUnit CustomAttributeTest](https://github.com/Scandltd/storm-petrel/blob/main/generator/Test.Integration.XUnit/CustomAttributes/CustomAttributeTest.cs), [RetryTest](https://github.com/Scandltd/storm-petrel/blob/main/generator/Test.Integration.XUnit/RetryTest.cs), [WpfCustomAttributeTest](https://github.com/Scandltd/storm-petrel/blob/main/generator/../file-snapshot-infrastructure/Test.Integration.WpfAppTest/WpfCustomAttributeTest.cs) tests and related custom attribute implementation files.

## Supported Software

### Test Frameworks
* [xUnit](https://xunit.net/)
    * **For v2 and v3**: See [Test.Integration.XUnit](https://github.com/Scandltd/storm-petrel/blob/main/generator/Test.Integration.XUnit) for examples.
    * **v3-specific**: See [Test.Integration.XUnitV3](https://github.com/Scandltd/storm-petrel/blob/main/generator/Test.Integration.XUnitV3).
* [NUnit](https://nunit.org/)
* [MSTest](https://github.com/microsoft/testfx/)

### .NET Versions
* .NET Standard 2.0+
* .NET 8.0+
* .NET Framework 4.6.2+

### Operating Systems and IDEs
* Windows, Linux. See more details about macOS and IDEs [here](#what-operating-systems-and-ides-can-i-use-to-run-stormpetrel-tests).

## CHANGELOG

See [CHANGELOG](https://github.com/Scandltd/storm-petrel/blob/main/generator/CHANGELOG.md) for more details.

## FAQs

### Does Scand.StormPetrel help developers track incorrect expected baselines in tests?
Developers can already track incorrect expected baselines without *Scand.StormPetrel*. They *should* manually review the changes made by *Scand.StormPetrel* to the expected baselines and decide if they are correct. This is the same approach used when *Scand.StormPetrel* is not involved.

### Does Scand.StormPetrel violate Test Driven Development (TDD) practices?
We believe it does not. Here are the corresponding TDD [steps](https://martinfowler.com/bliki/TestDrivenDevelopment.html) with comments explaining why it does not violate the practices:

* **Write a test for the next bit of functionality you want to add.**\
*Scand.StormPetrel* does *not* assist in creating test case input parameters.
It cannot help create the expected baseline at this step because there is no actual result yet. Therefore, the expected baseline is either:
    * A default instance
    * A manually created instance
* **Write the functional code until the test passes.**\
*Scand.StormPetrel* *can* automatically update the default or manually created instance with the actual value here. The developer *should* then review the changes to decide if the test truly passes.

### What is the suggested configuration of Scand.StormPetrel for typical CI/CD and development environments?
The suggested configuration in the `appsettings.StormPetrel.json` file is:
```jsonc
{
    "$schema": "https://raw.githubusercontent.com/Scandltd/storm-petrel/main/generator/assets/appsettings.StormPetrel.Schema.json", //Specify the path to observe schema suggestions in IDEs.
    "GeneratorConfig": {
        "BackuperExpression": null, //No need to backup because developers typically keep backups under Git or other Version Control Systems.
        "DumperExpression": ...,    //According to your requirements.
        "RewriterExpression": ...   //According to your requirements.
    },
    "IsDisabled": false,            //Keep `true` under Git control for CI/CD to speed up test compilation and execution time.
                                    //Keep `false` on the developer's machine, which should not be tracked by Git. StormPetrel tests are always available.
                                    //Keep `true` on the developer's machine. Can be changed ad hoc to `false` to compile StormPetrel tests.
    "Serilog": null                 //Avoid logging.
}
```

### Can test project compilation fail after enabling Scand.StormPetrel? How can I avoid the failure?
*Scand.StormPetrel* relies on code syntax, not semantics. Therefore, it cannot properly generate *StormPetrel* test methods in all cases, and test project compilation might fail. You can detect the original test file causing the failure and add it to the `"IgnoreFilePathRegex"` property in the [configuration](#configuration) to avoid the compilation error while still using *Scand.StormPetrel* for other tests.

### Scand.StormPetrel constantly changes my expected baseline because the actual result has new property values (e.g., CreatedOn with the current time, auto-incremented Id, random Guid, etc.) every StormPetrel test call. How can I avoid these constant changes?
It is likely that you also ignore the property in the test assertion; otherwise, the test would fail.
An option is to always have a default value for the property while dumping it to C# code using *Scand.StormPetrel*.
This can be implemented via custom configuration or the implementation of [IGeneratorDumper](https://github.com/Scandltd/storm-petrel/blob/main/generator/../abstraction/Scand.StormPetrel.Generator.Abstraction/IGeneratorDumper).
See an example of how this is implemented via the `GetDumpOptions` method in [Test.Integration.XUnit/Utils](https://github.com/Scandltd/storm-petrel/blob/main/generator/Test.Integration.XUnit/Utils.cs) and configured in [Test.Integration.XUnit/appsettings.StormPetrel.json](https://github.com/Scandltd/storm-petrel/blob/main/generator/Test.Integration.XUnit/appsettings.StormPetrel.json).

### What Operating Systems and IDEs can I use to run StormPetrel tests?
You can definitely use Windows or Linux with:
* Command line tools like `dotnet test ... --filter "FullyQualifiedName~StormPetrel"`.
* [Visual Studio](https://visualstudio.microsoft.com/) Test Explorer.

You can also likely use:
* macOS.
* Test runners in other IDEs like [Visual Studio Code](https://code.visualstudio.com/) or [JetBrains Rider](https://www.jetbrains.com/rider/) on Windows, Linux, or macOS, if applicable.

## References

At SCAND, we specialize in [building advanced .NET solutions](https://scand.com/technologies/net/) to help businesses develop new or modernize their legacy applications. If you need help getting started with Storm Petrel or support with implementation, we're ready to assist. Whether you're refactoring or rewriting, our team can help solve any challenges you might face. Visit our [page](https://scand.com/contact-us/) to learn more, or reach out for hands-on guidance.